#include "tokenstype.h"

enum TokenClass {
    BUILT_IN_FUNC = 1,
    DATA_TYPE     = 2,
    KEYWORD_NAME  = 4,
    TOKENIZABLE   = 8
};


struct TokenInfo {
    std::string text;
    unsigned short attributes;
};



struct TokenTables {
    std::unordered_map<TokenType, TokenInfo> allTokensInfo;
    std::unordered_map<std::string, TokenType> keywords;

    TokenType tokenize(std::string value, TokenType initial);

    void add(TokenType tokenType, const std::string& txt,
             unsigned short attribs = 0);

    TokenTables();
};


static TokenTables tokenTables;

void TokenTables::add (TokenType tokenType, const std::string& txt,
                        unsigned short attribs)
{
    allTokensInfo[tokenType] = TokenInfo{txt, attribs};
    if(attribs & KEYWORD_NAME)
        keywords[txt] = tokenType;
}

TokenTables::TokenTables() {

    add( TokenType::IDENTIFIER, "identifier" );
    add( TokenType::KEYWORD,    "keyword" );

    // Literals (bool is not here, it has two keywords: false and true)
    add( TokenType::NUMBER,     "number");
    add( TokenType::STRING,     "string");
    add( TokenType::COLOR,      "color");

    // Parentheses of all kinds
    add( TokenType::LPAREN,     "(" );
    add( TokenType::RPAREN,     ")" );
    add( TokenType::LBRACKET,   "[" );
    add( TokenType::RBRACKET,   "]" );
    add( TokenType::LBRACE,     "{" );
    add( TokenType::RBRACE,     "}" );
    add( TokenType::LT,         "<" );
    add( TokenType::GT,         ">" );

    // assignments
    add( TokenType::EQUALS,     "=" );

    // miscellaneous
    add( TokenType::DOT,        "." );
    add( TokenType::COLON,      ":" );
    add( TokenType::SEMICOLON,  ";" );
    add( TokenType::COMMA,      "," );
    add( TokenType::WHITESPACE, "whitespace" );
    add( TokenType::COMMENT,    "comment" );

    add( TokenType::TRUE, "true", TOKENIZABLE ),
    add( TokenType::FALSE, "false", TOKENIZABLE),

    add( TokenType::IF, "if", TOKENIZABLE ),
    add( TokenType::ELSE, "else", TOKENIZABLE),

    add( TokenType::DEFAULT, "default", TOKENIZABLE),
    add( TokenType::NIL, "null", TOKENIZABLE),

    // all keywords
    add( TokenType::KW_NAME, "@name", TOKENIZABLE),
    add( TokenType::KW_INFO, "@info", TOKENIZABLE),
    add( TokenType::KW_HEADER, "@header", TOKENIZABLE),
    add( TokenType::KW_DIRECTORY, "@directory", TOKENIZABLE),
    add( TokenType::KW_LIBRARY, "@library", TOKENIZABLE),
    add( TokenType::KW_ADDRESS, "@address", TOKENIZABLE),
    add( TokenType::KW_COMPONENT, "@component", TOKENIZABLE),
    add( TokenType::KW_MESSAGES, "@messages", TOKENIZABLE),
    add( TokenType::KW_INSTANCE_NAME, "@instanceName", TOKENIZABLE),
    add( TokenType::KW_COUNT, "@count", TOKENIZABLE),
    add( TokenType::KW_DISPLAY, "@display", TOKENIZABLE),
    add( TokenType::KW_PIN, "@pin", TOKENIZABLE),
    add( TokenType::KW_TOOLTIP, "@tooltip", TOKENIZABLE),
    add( TokenType::KW_CONNECTION, "@connection", TOKENIZABLE),
    add( TokenType::KW_ATTRIBUTE, "@attribute", TOKENIZABLE),
    add( TokenType::KW_SOURCE, "@source", TOKENIZABLE),
    add( TokenType::KW_POPUP, "@popup", TOKENIZABLE),
    add( TokenType::KW_RULE, "@rule", TOKENIZABLE),
    add( TokenType::KW_TITLE, "@title", TOKENIZABLE),
    add( TokenType::KW_TEXT, "@text", TOKENIZABLE),
    add( TokenType::KW_BUS, "@bus", TOKENIZABLE),
    add( TokenType::KW_WIRES, "@wires", TOKENIZABLE),
    add( TokenType::KW_ENUMERATED, "@enumerated", TOKENIZABLE),
    add( TokenType::KW_WIRE, "@wire", TOKENIZABLE),
    add( TokenType::KW_INSTANCE, "@instance", TOKENIZABLE),
    add( TokenType::KW_SCHEMA, "@schema", TOKENIZABLE),
    add( TokenType::KW_POSITION, "@position", TOKENIZABLE),
    add( TokenType::KW_SIZE, "@size", TOKENIZABLE),

    // All types
    add( TokenType::INT_TYPE, "int", TOKENIZABLE),
    add( TokenType::STRING_TYPE, "string", TOKENIZABLE),
    add( TokenType::BOOL_TYPE, "bool", TOKENIZABLE),
    add( TokenType::WIRE_TYPE, "wire", TOKENIZABLE),

    // WireNode types
    add( TokenType::WIRE_DEFAULT, "normal_wire", TOKENIZABLE),
    add( TokenType::WIRE_AND, "wired_and", TOKENIZABLE),
    add( TokenType::WIRE_OR, "wired_or", TOKENIZABLE),
    add( TokenType::R_WIRE, "r_wire", TOKENIZABLE),

    // ComponentNode types
    add( TokenType::CT_MEMORY, "memory", TOKENIZABLE),
    add( TokenType::CT_PROCESSOR, "processor", TOKENIZABLE),

    // PinNode type
    add( TokenType::PIN_IN, "in", TOKENIZABLE),
    add( TokenType::PIN_OUT, "out", TOKENIZABLE),
    add( TokenType::PIN_IN_OUT, "inOut", TOKENIZABLE),

    // ActionNode types
    add( TokenType::ERROR, "error", TOKENIZABLE),
    add( TokenType::WARNING, "warning", TOKENIZABLE),


    // Built-in functions (they are also keywords)
    /*
     * TODO
    */
    // Miscelaneous
    add( TokenType::END_OF_FILE,   "eof" );
}


const std::string &tokenTypeToString(TokenType tokenType) {
    return tokenTables.allTokensInfo[tokenType].text;
}

TokenType TokenTables::tokenize(std::string value, TokenType initial) {
    for(auto& [key, param]: allTokensInfo) {
        if(param.attributes & TOKENIZABLE && param.text == value) {
            return key;
        }
    }
    return initial;
}

TokenType from_token(std::string value, TokenType initial) {
    return tokenTables.tokenize(value, initial);
}